/* protocol.c       Copyright (c) 2000-2002 Nagy Daniel
 *
 * $Date: 2002/06/19 14:09:35 $
 * $Revision: 1.12 $
 *
 * This module deals with incoming and outgoing packets.
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version, with the
 * additional permission that it may be linked against Erick Engelke's
 * WATTCP source code and Jerry Joplin's CVT100 source code.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <mem.h>

#if defined (__DJGPP__)
 #include <errno.h>
 #include <string.h>
 #include "tcp_djgp.h"
#elif defined (__TURBOC__)
 #include "tcp.h"
#endif

#include "ssh.h"
#include "protocol.h"
#include "cipher.h"
#include "crc.h"
#include "zlib.h"
#include "xmalloc.h"
#include "macros.h"
#include "config.h"

void fatal(const char *, ...);

extern Config GlobalConfig;
extern ConfigBits GlobalConfigBits;
extern int status;

struct Packet pktin = { 0, 0, 0, NULL, 0 };
struct Packet pktout = { 0, 0, 0, NULL, 0 };

static z_stream out_stream;	/* outgoing compressed stream */
static z_stream in_stream;	/* incoming stream to uncompress */

CryptPtr EncryptPacket;
CryptPtr DecryptPacket;

char *BadPacket="Received invalid packet";
char *ConnectionClosed="Connection closed";
char *RemoteClosed="Remote host closed connection";

/* request compression in protocol negotiation phase. */
void Request_Compression(int level)
{
char nomem[]="Not enough memory for compression support";

   if(GlobalConfigBits.verbose) cputs("Requesting compression\r\n");

   s_wrpkt_start(SSH_CMSG_REQUEST_COMPRESSION, 4);
   pktout.body[0] = pktout.body[1] = pktout.body[2] =0;
   pktout.body[3] = level; /* compression level */
   s_wrpkt();
   packet_read_block();
   switch (pktin.type){
	case SSH_SMSG_SUCCESS:
	   if((deflateInit(&out_stream, level))==Z_MEM_ERROR)
			fatal(nomem);
	   if((inflateInit(&in_stream))==Z_MEM_ERROR)
			fatal(nomem);
	   GlobalConfigBits.Compressed = 1;
	   break;

	case SSH_SMSG_FAILURE:
	   cputs("Server refused to compress\r\n");
	   break;

	default:
	   fatal(BadPacket);
	} /* switch */
}

/* clean up compression structures at the end. This frees memory */
void Uninit_Compression(void)
{
   deflateEnd(&out_stream);
   inflateEnd(&in_stream);
}


/* compress a buffer */
int sshcompress (unsigned char *dest, unsigned long *destLen,
		 unsigned char *source, unsigned long sourceLen)
{
int err;

   out_stream.next_in = (Bytef*)source;		/* source buffer */
   out_stream.avail_in = (uInt)sourceLen;	/* source length */
   out_stream.next_out = (Bytef*)dest;		/* destination buffer */
   out_stream.avail_out = (uInt)*destLen;	/* destination buffer size */

   err = deflate(&out_stream, Z_PARTIAL_FLUSH);/* we do the whole at once,
						   because buffer is large */
   *destLen = *destLen - out_stream.avail_out;	/* return compressed size */

   return err;
}

/* uncompress a buffer */
int sshuncompress (unsigned char *dest, unsigned long *destLen,
		   unsigned char *source, unsigned long sourceLen)
{
int err;

   in_stream.next_in = (Bytef*)source;
   in_stream.avail_in = (uInt)sourceLen;
   in_stream.next_out = (Bytef*)dest;
   in_stream.avail_out = (uInt)*destLen;

   err = inflate(&in_stream, Z_PARTIAL_FLUSH);

   *destLen = *destLen - in_stream.avail_out;

   return err;
}


/* save packets in hexadecimal format for debugging */
static void fwritehex(unsigned char *buf, unsigned int length)
{
unsigned int i;

   for(i = 1; i <= length; i++){
	fprintf(GlobalConfig.debugfile, "%02X ", *buf++);
	if(i%16==0)
	   fputs("\n", GlobalConfig.debugfile);
   }

   fputs("\n", GlobalConfig.debugfile);
   buf-=length;

   for(i = 1; i <= length; i++){
	if(*buf >= ' ' && *buf < 126)
	   fprintf(GlobalConfig.debugfile, "%c", *buf);
	else
	   fputc('.', GlobalConfig.debugfile);
	if(i%16==0)
	    fputs("\n", GlobalConfig.debugfile);
	buf++;
   }
}


/* Read and convert raw packet to readable structure.
 * Uncrypt and uncompress if necessary */
void ssh_gotdata(void)
{
unsigned long len;	/* packet length without length and padding fields */
unsigned long PktLength;/* full packet length with padding */
unsigned int pad;	/* number of padding bytes */
unsigned char PktInLength[4];	/* first four bytes of a packet (length) */
unsigned char *inbuf;           /* buffer for incoming packet */
unsigned char *decompblk;	/* buffer for decompression */
unsigned long decomplen;	/* Entry: buffer size. Exit: data langth */
unsigned short i;

/*
1. Get four bytes from packet, which is the packet size without padding.
2. Calculate the padding length and full packet length.
3. Allocate memory for the whole packet and read it.
4. Uncrypt packet.
5. Fill incoming packet structure
6. Verify CRC (padding + type + body)
7. Uncompress packet (type + body)
8. Get packet type
*/

   sock_fastread(&GlobalConfig.s, PktInLength, 4);
   for (i = len = 0; i < 4; i++)
       len = (len << 8) + PktInLength[i];
   pad = 8 - (len%8);
   PktLength = len + pad;
   inbuf = (unsigned char *)xmalloc(PktLength * (sizeof(char)));
   sock_read(&GlobalConfig.s, inbuf, PktLength); /* Read rest */

   if(GlobalConfigBits.cipher) /* uncrypt */
		DecryptPacket(inbuf, PktLength);

   pktin.length = PktLength - 4 - pad;	/* minus CRC, padding */
   pktin.whole = (pktin.whole == NULL ? (unsigned char *)xmalloc(pktin.length) :
			(unsigned char *)xrealloc(pktin.whole, pktin.length));
   memcpy(pktin.whole, inbuf + pad, pktin.length);
   xfree(inbuf); /* it's now in pktin structure, so free it */

   if (GlobalConfigBits.Compressed){
	i = MAX_PACKET_SIZE;
	decomplen = i*5; /* we shouldn't receive larger...*/
	decompblk = (unsigned char *)xmalloc(decomplen);
	if((i = sshuncompress(decompblk, &decomplen, pktin.whole,
		pktin.length)) != Z_OK ) fatal("Decompression error: %d\n", i);
	pktin.whole = (unsigned char *)xrealloc(pktin.whole, decomplen);
	memcpy(pktin.whole, decompblk, decomplen); /* copy uncompressed */

	xfree(decompblk);
	pktin.length = decomplen;	/* fix length after uncompression */
    } /* Compression */

   pktin.body = pktin.whole + 1; /* skip type field */
   pktin.type = *pktin.whole;    /* first byte is the type */

   if(GlobalConfigBits.debug){
	fputs("\nRECEIVED packet:\n", GlobalConfig.debugfile);
	fwritehex(pktin.whole, pktin.length);
	fputc('\n', GlobalConfig.debugfile);
   } /* debug */

}

/* calculate full packet length from body length */
void s_wrpkt_size(unsigned long len)
{
unsigned short pad;
unsigned long PktLength;

   len += 5;			/* add type and CRC */
   pad = 8 - (len%8);           /* calculate padding */
   PktLength = len + pad + 4;	/* add padding and length fileld for total length */

   pktout.length = len - 4;
   pktout.whole = (pktout.whole == NULL ? (unsigned char *)xmalloc(PktLength) :
		       (unsigned char *)xrealloc(pktout.whole, PktLength));
   pktout.body = pktout.whole + 5 + pad;
}

/* create header for raw outgoing packet */
void s_wrpkt_start(int type, int len)
{
   s_wrpkt_size(len);
   pktout.type = type;
}

/* create outgoing packet */
void s_wrpkt(void)
{
int pad, len, PktLength, i;
unsigned long crc;
unsigned char *compblk;
unsigned long complen;

/*
1. Add type
2. Compress (type + data)
3. Calculate and add padding
4. Calculate length (type + body + CRC)
5. Calculate CRC (padding + type + body)
6. Encrypt (padding + type + data + CRC)
7. Send
*/

   pktout.body[-1] = pktout.type;

   if (GlobalConfigBits.Compressed) {
	complen=(pktout.length)+(pktout.length)/1000+13; /* from zlib */
	compblk=(unsigned char *)xmalloc(complen);
	if((i = sshcompress(compblk, &complen, pktout.body - 1, pktout.length))
				 != Z_OK) fatal("Compression error: %d\n", i);
	s_wrpkt_size(complen - 1);
	memcpy(pktout.body - 1, compblk, complen);
	xfree(compblk);
   } /* Compression */

   len = pktout.length + 4;	       /* plus CRC length */
   pad = 8 - (len%8);
   PktLength = len + pad;

   for (i = 0; i < pad; i++)		/* add padding */
	pktout.whole[i+4] = rand() % 256;

   crc = sshcrc32(pktout.whole + 4, PktLength - 4);

   PUT_32BIT_MSB_FIRST(pktout.whole + PktLength, crc);
   PUT_32BIT_MSB_FIRST(pktout.whole, len);

   if(GlobalConfigBits.debug){
	fputs("\nSENT packet:\n", GlobalConfig.debugfile);
	fwritehex(pktout.body - 1, pktout.length);
	fputc('\n', GlobalConfig.debugfile);
   }

   if(GlobalConfigBits.cipher)
	EncryptPacket(pktout.whole + 4, PktLength);

   if(sock_write(&GlobalConfig.s, pktout.whole, PktLength + 4) != PktLength + 4)
	        fatal("write: %s", strerror(errno));
}

/* Get a packet with blocking. Handle debug, ignore and disconnect packets */
int packet_read_block(void)
{
restart:

   sock_wait_input(&GlobalConfig.s, sock_delay, NULL, &status);
   ssh_gotdata();

   if(pktin.type==SSH_MSG_DEBUG){
	if(GlobalConfigBits.debug || GlobalConfigBits.verbose){
	   pktin.body[4+pktin.body[3]]=0;
	   printf("DEBUG: %s\n",pktin.body+4);
	}
        goto restart;
   }

   if(pktin.type==SSH_MSG_IGNORE)
	goto restart;

   if(pktin.type==SSH_MSG_DISCONNECT){
	pktin.body[4+pktin.body[3]]=0;
        fatal("DISCONNECT: %s\n",pktin.body+4);
   }

return(0);

sock_err:
   switch (status){
	case 1 :
	   cputs (ConnectionClosed);
           break;

	case -1:
	   cputs (RemoteClosed);
           break;
  }
   fatal("Socket error");
   return(0);	/* for BorlandC */
}

/* expect a packet type */
void packet_read_expect(int type)
{
   packet_read_block();
   if(pktin.type!=type)
	fatal(BadPacket);
}


/* Send n data bytes as an SSH packet */
void SendSSHPacket(char *buff, unsigned short len)
{
   s_wrpkt_start(SSH_CMSG_STDIN_DATA,len + 4);
   pktout.body[0]=(len & 0xff000000) >> 24;
   pktout.body[1]=(len & 0xff0000) >> 16;
   pktout.body[2]=(len & 0xff00) >> 8;
   pktout.body[3]= len & 0xff;
   memcpy(&pktout.body[4], buff, len);
   s_wrpkt();
}
